#######################################################################
###### Plugin Software was Designed and Developed by
###### BDBI, Product Innovation.
###### Authored by Joshua Luthy
###### This template was last updated in May 2023
#######################################################################
## Just for debugging to see what FlowJo is actually passing to the script.
##
## cat("FJ-DATA-FILE-PATH            = FJ_SAMPLE_DATA_PATH \n")
## cat("FJ-CSV-OUPUT-DIR             = FJ_OUTPUT_DIRECTORY_PATH \n")
## cat("FJ-PAR-NUM-CLUST             = FJ_PAR_NUM_CLUST \n")
## cat("FJ-METHODSFILE-PATH          = FJ_METHODSFILE_PATH \n")
## cat("FJ-PARAMS-LIST               = FJ_PARAMS_LIST \n")
## cat("FJ-BG-META-CLUSTERS          = FJ_BG_META_CLUSTERS \n")
## cat("FJ-PLOT-OPTION               = FJ_PLOT_OPTION \n")
## cat("FJ-VIEW-OPTION               = FJ_VIEW_OPTION \n")
## cat("FJ-PLOT-ALL                  = FJ_PLOT_ALL \n")
## cat("FJ-PAR-SCALE                 = FJ_PAR_SCALE \n")
## cat("FJ-PAR-XDIM                  = FJ_PAR_XDIM \n")
## cat("FJ-PAR-YDIM                  = FJ_PAR_YDIM \n")
## cat("FJ-PAR-CREATE-PDF            = FJ_PAR_CREATE_PDF \n")
## cat("FJ-PAR-NODE-SCALE            = FJ_PAR_NODE_SCALE \n")
## cat("FJ-PAR-FIXED-NODE-SIZE       = FJ_PAR_FIXED_NODE_SIZE \n")
## cat("FJ-PAR-APPLY-ON-PREV         = FJ_PAR_APPLY_ON_PREV \n")
## cat("FJ-PAR-ADD-CELLIDS-TO-RESULT = FJ_PAR_ADD_CELLIDS_TO_RESULT \n")
## cat("FJ-OUTS-FOLDER               = FJ_OUTS_FOLDER \n")
## cat("FJ-PAR-CREATE-SOMMARY        = FJ_PAR_CREATE_SOMMARY \n")
## cat("FJ-PLOT-BKGD                 = FJ_PLOT_BKGD \n")
## cat("FJ-PLOT-RADAR                = FJ_PLOT_RADAR \n")
## cat("FJ-SEED                      = FJ_SEED \n")
## cat("FJ-POPNAME                   = FJ_POPNAME \n")
#######################################################################
##         LOAD INPUTS  & FUNCTIONS     ##
#######################################################################
outPutFolder <- "FJ_OUTPUT_DIRECTORY_PATH"
inputCSVpath <- "FJ_SAMPLE_DATA_PATH"
applyOnMapPath <- "FJ_PAR_APPLY_ON_PREV"
applyOnMap <- FJ_HAS_MAP
runID <- "FJ_RUNID"
supLabel <- "FJ_SUPER_LABEL"
supLabelbkup <- "FJ_SUP_LABEL_BKUP"
fjPopName <- "FJ_POPNAME"
if (supLabel == "fsupervisedID") {
  supLabel <- supLabelbkup
}
isSeqGeq <- FJ_IS_SEQGEQ
# UMAP
prevFJobj <- NULL
fDistMetric <- "FJ_DIST_METRIC"
fNneighbors <- FJ_N_NEIGHBORS
fNumComponents <- FJ_NUM_COMPONENTS
fMinDist <- FJ_MIN_DIST
runSUMAP <- FJ_RUN_SUPER_UMAP
fSuperWeight <- FJ_SUPER_WEIGHT
fjSeed <- FJ_SEED
# Constants
EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"

source("FJ_METHODSFILE_PATH")

StopAndReturnError <- function(message, ...) {
  cat(message, file=paste0(outPutFolder, "Error_Message_from_R.txt"))
  stop(message, ...)
}
#######################################################################
###### Start package import or installations
#######################################################################
# Install function for packages
loadPackages <-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    install.packages(need, repos="https://cran.r-project.org")
    lapply(need,require,character.only=TRUE)
  }
}

loadBioc<-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    BiocManager::install(need)
    lapply(need,require,character.only=TRUE)
  }
}
tryCatch({
  loadPackages(c('ggplot2', 'FNN', 'igraph', 'Matrix', 'cowsay', 'umap', 'uwot', 'utils', 'devtools', 'data.table', 'dplyr'))
}, error = function(e) {
  message <- "There was an error with installing R packages. Please see How-To PDF and try to manually install."
  StopAndReturnError(message = message)
}, warning = function(w){
  loadPackages(c('ggplot2', 'FNN', 'igraph', 'Matrix', 'cowsay', 'umap', 'uwot', 'utils', 'devtools', 'data.table', 'dplyr'))
  writeLines("loaded R packages")
})

sessionInfo()

if(applyOnMap){
  ## Apply on Map
  mapFileExtension <- tools::file_ext(applyOnMapPath)
  if (mapFileExtension == "RDS") {
    tryCatch({
      umapModel <- readRDS(applyOnMapPath)
    }, error = function(e){
      message <- "Cannot load the saved object. Verify the path or try loading it from an alternate location."
      StopAndReturnError(message = message)
    })
  }
}
if(length(applyOnMapPath) > 5) {
  print("Run new UMAP embedding")
} else {
  print("Apply on Previous UMAP model")
}
################################################################
# Create the basic FlowJo object ad fill slots as we go
################################################################
myFJobj <- new("FJObj", runID = runID)
################################################################
###### Start the File Import
################################################################
myFJobj <- loadFlowJoCSV(myFJobj, inputCSVpath)
###############################################################
#  Create export object   #
###############################################################
if(isSeqGeq){
  r1 <- myFJobj@inputCSV$cellIds
} else {
  r1 <- myFJobj@inputCSV$eventNumbers
}
###############################################################
#  Dim Reduction with UMAP   #
###############################################################
set.seed(fjSeed) # all 3 map types below require random number

if(applyOnMap){
  #TODO perform some additional checks here to ensure matching parameters are present in both selected pop and previous map or return error text file
  compareNewColumnNames <- function(df1, df2) {
    commonColumnNamz <- intersect(colnames(df1), colnames(df2))
    return(commonColumnNamz)
  }
  commonCols <- compareNewColumnNames(myFJobj@inputCSV$X, umapModel@inputCSV$X)

  savedModel <- load_uwot(umapModel@EmbedSOMReduction$map, verbose = TRUE)
  myFJobj@EmbedSOMReduction$embeddingMat <- uwot::umap_transform(X=myFJobj@inputCSV$X[ , commonCols], model=savedModel)

} else {
  if(runSUMAP){
    print("Performing Supervised embedding")
    tempMat <- as.data.frame(myFJobj@inputCSV$X)
    clusterMat <- as.matrix(tempMat[supLabel])
    clusterMat <- dplyr::na_if(clusterMat, 0)
    clusterMat <- as.factor(clusterMat)
    print(levels(clusterMat))

    embedding <- uwot::umap(X=myFJobj@inputCSV$X, n_components = fNumComponents, metric = fDistMetric, min_dist = fMinDist,
                            n_neighbors = fNneighbors, y = clusterMat,ret_model=TRUE)
    save_uwot(embedding, file = paste0(outPutFolder, "UMAP_Model_", runID), unload = FALSE, verbose = TRUE)
    myFJobj@EmbedSOMReduction$embeddingMat <- embedding$embedding
    myFJobj@EmbedSOMReduction$map <- paste0(outPutFolder, "UMAP_Model_", runID)
    derParName <- paste0("sUMAP")
    myFJobj@EmbedSOMReduction$embeddingType <- 'sUMAP'

  } else {
    print("Performing UMAP embedding")
    embedding <- uwot::umap(X=myFJobj@inputCSV$X, n_components = fNumComponents, metric = fDistMetric, min_dist = fMinDist,
                            n_neighbors = fNneighbors, ret_model=TRUE)
    save_uwot(embedding, file = paste0(outPutFolder, "UMAP_Model_", runID), unload = FALSE, verbose = TRUE)
    myFJobj@EmbedSOMReduction$embeddingMat <- embedding$embedding
    myFJobj@EmbedSOMReduction$map <- paste0(outPutFolder, "UMAP_Model_", runID)
    derParName <- paste0("UMAP")
    myFJobj@EmbedSOMReduction$embeddingType <- 'UMAP'
  }
}


# embedding is ready to be sent back to FlowJo.
if(applyOnMap){
  modelRunID <- getRunID(umapModel)
  modelInfo <- getEmbeddingInfo(umapModel)
  colnames(myFJobj@EmbedSOMReduction$embeddingMat) <- paste0(modelInfo$type, ".", 1:modelInfo$nDims, ".",fjPopName,".",modelRunID)
} else {
  colnames(myFJobj@EmbedSOMReduction$embeddingMat) <- paste0(derParName, ".", 1:fNumComponents, ".",fjPopName,".",runID)
}

if ((NA %in% myFJobj@EmbedSOMReduction$embeddingMat) || (NaN %in% myFJobj@EmbedSOMReduction$embeddingMat)) {
  stopAndReturnError("Unfortunately, we were unable to calculate EmbdedSOM results for this input\n(NaN in the resulting matrix). You may want to change the input arguments.")
} else {
  r1 <- cbind(r1, myFJobj@EmbedSOMReduction$embeddingMat)
}

###############################################################
#  Exporting of results to a csv file to be merged to FJ   #
###############################################################
write.csv(r1, file=paste0(outPutFolder, "UMAP_Results_", runID, ".csv"), row.names=FALSE, quote=FALSE)
if(!applyOnMap){
  saveRDS(myFJobj, file = paste0(outPutFolder, "UMAP_RData_", runID, ".RDS"), compress = FALSE)
}
rm(list = ls())
say("time")